/*
	EASY-ROB Development V2.3

	12 user defined inverse kinamatic routines are suported.
	int inv_kin_user_1..12 (FRAME *)

	12 user defined functions for mathematical joint dependencies are suported.
	float md_f1..12 (void)

	The EASY-ROB base program "easyrobw.exe" provides more than 100 functions.

	Stefan Anton

	modul: inv_user.cpp
*/
#include "stdafx.h"

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "..\er_dvlp\er_dvlp.h"

#pragma warning(disable : 4244)     // float conversation

#define NUM_DOFS_1 6
#define NUM_CONFIGS_1 8

#define NUM_DOFS_2 6
#define NUM_CONFIGS_2 8

#define NUM_DOFS_3 9
#define NUM_CONFIGS_3 8

#define NUM_DOFS_4 6
#define NUM_CONFIGS_4 8

#define NUM_DOFS_5 6
#define NUM_CONFIGS_5 8

#define NUM_DOFS_6 6
#define NUM_CONFIGS_6 8

#define NUM_DOFS_7 6
#define NUM_CONFIGS_7 8

#define NUM_DOFS_8 6
#define NUM_CONFIGS_8 8

#define NUM_DOFS_9 6
#define NUM_CONFIGS_9 8

#define NUM_DOFS_10 6
#define NUM_CONFIGS_10 8

#define NUM_DOFS_11 6
#define NUM_CONFIGS_11 8

#define NUM_DOFS_12 6
#define NUM_CONFIGS_12 8

DllExport int er_kin(void)
{
	return(0);
}
DllExport int  inv_kin_user_2(frame *bTt)
{
  double *achs_length,*achs_offsets1,*achs_offsets2,
		 *sing_tol,*joint_offset;
  float *q;

  int *warnings,*verpolt;
  int num_dofs,num_configs;
  frame *T,b0Tt;

  int i,ii, jj, kk, soln, warning;
  double nx,ny,nz,ox,oy,oz,ax, ay, az,px, py, pz;
  float l1x, l1y, l1z, l2x, l2y, l2z, l2, l3x, l3y, l3z, l3, l6,
		 ly, rx, rxy, rxm, rr;
  double c1, s1, c2, c3, c23, s23,s6,c6,s4,c4, cabs;
  double theta[KIN_DOFS];
  double delta_theta1, delta_theta2, delta_theta3;
  float sig1, sig2, sig3, sing_tolx, sing_tol_q5;

//rob_kin_user_msg("Hallo, try this one");

	 num_dofs = inq_num_dofs();
	 num_configs = inq_num_configs();

	 if (num_dofs!=NUM_DOFS_2 || num_configs!=NUM_CONFIGS_2) {
		 char b[80];
		 rob_kin_user_msg("ERROR inv_kin_user_2()");
		 sprintf(b,"dofs    %2d != %2d",num_dofs,NUM_DOFS_2);
		 rob_kin_user_msg(b);
		 sprintf(b,"configs %2d != %2d",num_configs,NUM_CONFIGS_2);
		 rob_kin_user_msg(b);
	 }

	 achs_length = inq_achs_length();
	 achs_offsets1 = inq_achs_offsets1();
	 achs_offsets2 = inq_achs_offsets2();
	 joint_offset = inq_joint_offset();
	 sing_tol = inq_sing_tol();
	 warnings = inq_warnings();
	 verpolt = inq_verpolt();

	 l1z = achs_length[0];
	 l1x = achs_offsets1[0];
	 l1y = achs_offsets2[0];

	 l2z = achs_length[1];
	 l2x = achs_offsets1[1];
	 l2y = achs_offsets2[1];

	 if(fabs(l2x)>0.0 && fabs(l2z)>0.0) {
		 delta_theta2 = atan(l2x/l2z);
		 l2  = sqrt( l2x*l2x + l2z*l2z );
	 } else {
		 delta_theta2 = 0.0;
		 l2 = l2z;
	 }

	 l3z = achs_length[2];
	 l3x = achs_offsets1[2];
	 l3y = achs_offsets2[2];

	 if(fabs(l3x)>0.0 && fabs(l3z)>0.0) {
		 delta_theta3 = atan(l3x/l3z);
		 l3  = sqrt( l3x*l3x + l3z*l3z );
	 } else {
		 delta_theta3 = 0.0;
		 l3 = l3z;
	 }

	 ly = l1y + l2y + l3y;

	 l6 = achs_length[5];

	 sing_tolx = sing_tol[0]; /* [mm] */
	 sing_tol_q5 = (float)sin(sing_tol[4]); /* [rad] */

	 T = &b0Tt;
	 mul_invT_T(T,inq_bT0(),bTt);

	 px = T->p[0]-T->a[0]*l6;
	 py = T->p[1]-T->a[1]*l6;
	 pz = T->p[2]-T->a[2]*l6-l1z;

	 rxy = sqrt(px*px+py*py);

	 if( rxy < ly ) {
		  rob_kin_set_warnings(WARN_UNREACH);
		  return( WARN_UNREACH );
	 } else if( rxy < sing_tolx ) {
		  rob_kin_set_warnings(WARN_SINGULAR);
	 }

	 if (fabs(ly)>0.0) {
		 rx = rxy*rxy - ly*ly;
		 if (rx<0.0) rx=0.0;
		 else rx = sqrt( rx );
		 delta_theta1 = rob_kin_atan2(ly,rx);
	 } else {
		 rx = rxy;
		 delta_theta1 = 0.0;
	 }

	 sig1 = -1.0;
	 sig2 = -1.0;
	 soln = 0;


	for( kk = 0; kk < 2; kk++ )
	{

	 warning=WARN_OK;

	 sig1 *= -1.0;

	 rxm = rx - sig1*l1x;

	 rr = sqrt( rxm*rxm + pz*pz );

	 if (rr<KIN_EPSI) {
		  rob_kin_set_warnings(WARN_UNREACH);
		  return( WARN_UNREACH );
	 }

	 c2 = (rr*rr + l2*l2 - l3*l3) / (2.0*l2*rr);

	 if( (cabs=fabs(c2)) > 1.0 ) {
		if (cabs >1.0+KIN_EPSI) {
		  warning = WARN_UNREACH;
		}
		if (c2>1.0) c2 = 1.0;
		else c2 = -1.0;
	 }

	 c3 = (l3*l3 + l2*l2 - rr*rr) / (2.0*l2*l3);

	 if( (cabs=fabs(c3)) > 1.0 ) {
		if (cabs >1.0+KIN_EPSI) {
		  warning = WARN_UNREACH;
		}
		if (c3>1.0) c3 = 1.0;
		else c3 = -1.0;
	 }

	if (warning==WARN_OK) {
		theta[0] = -sig1*delta_theta1 +
					 rob_kin_atan2(sig1*py,sig1*px);
	}
	else {
		for (i=kk*4;i<(kk+1)*4;i++) warnings[i] = WARN_UNREACH;
		soln = (kk+1)*4;
	}

	for( jj = 0; jj < 2 && warning==WARN_OK; jj++ ) {

		 sig2 *= -1.0;

		 theta[1] = sig1*(PI/2.0-rob_kin_atan2(pz,rxm)-sig2*acos(c2))
					 - delta_theta2;

		 if (theta[1] < -PI) theta[1] += 2*PI;
		 else if (theta[1] > PI)  theta[1] -= 2*PI;

		 theta[2] = -sig1*sig2*acos( c3 )
					 - delta_theta3 + delta_theta2;

		 if( theta[2] < 0.0 ) theta[2] += PI;
		 else theta[2] -= PI;

		 c1 = cos( theta[0] );
		 s1 = sin( theta[0] );

		 c23 = cos( theta[2] + theta[1] );
		 s23 = sin( theta[2] + theta[1] );

		 nx = (T->n[0]*c1+T->n[1]*s1)*c23-T->n[2]*s23;
		 ny = -T->n[0]*s1+T->n[1]*c1;
		 nz = (T->n[0]*c1+T->n[1]*s1)*s23+T->n[2]*c23;

		 ox = (T->o[0]*c1+T->o[1]*s1)*c23-T->o[2]*s23;
		 oy = -T->o[0]*s1+T->o[1]*c1;
		 oz = (T->o[0]*c1+T->o[1]*s1)*s23+T->o[2]*c23;

		 ax = (T->a[0]*c1+T->a[1]*s1)*c23-T->a[2]*s23;
		 ay = -T->a[0]*s1+T->a[1]*c1;
		 az = (T->a[0]*c1+T->a[1]*s1)*s23+T->a[2]*c23;

		 sig3 = -1.0;

		 for( ii = 0; ii < 2 ; ii++ ) {

			 sig3 *= -1.0;

			 theta[5] = rob_kin_atan2(  sig3*oz,-sig3*nz);
			 c6 = cos(theta[5]); s6 = sin(theta[5]);
			 theta[3] = rob_kin_atan2(-ox*c6-nx*s6,oy*c6+ny*s6);
			 c4 = cos(theta[3]); s4 = sin(theta[3]);
			 theta[4] = rob_kin_atan2(ax*c4+ay*s4,az);

			 if( fabs( sin(theta[4]) ) < sing_tol_q5 )
			 {
				 rob_kin_user_msg("inv_kin_user_2 WARN_SINGULAR");
				 warnings[soln] = WARN_SINGULAR;
			 }
			 else warnings[soln] = WARN_OK;
//theta[5]=0;
			 for( i = 0; i < NUM_DOFS_2; i++ ) {
				 q = inq_q(soln);
				 q[i] =  verpolt[i]*(theta[i]
										+joint_offset[i]);

				 if (q[i] < -PI) q[i] += 2*PI;
				 else if (q[i] > PI)  q[i] -= 2*PI;
			 }
			 soln++;
		 } /* for ii */
	} /* for jj */
	} /* for kk */
	return(WARN_OK);
} /* End of inv_kin_user_2*/

DllExport int  inv_kin_user_1(frame *T)
{
	char b[256];
	double *v;
	int i,n;

	n=inq_num_dofs();

	if(n>KIN_DOFS) return(WARN_UNREACH);

	moni_v("kpp",inq_cntrl_kpp(),n,1);

	sprintf(b,"achs_off1");
	v = inq_achs_offsets1();
	for (i=0;i<n;i++) sprintf(&b[strlen(b)]," %10.4lf",v[i]);
	rob_kin_user_msg(b);

	sprintf(b,"achs_off2");
	v = inq_achs_offsets2();
	for (i=0;i<n;i++) sprintf(&b[strlen(b)]," %10.4lf",v[i]);
	rob_kin_user_msg(b);

	sprintf(b,"achs_len ");
	v = inq_achs_length();
	for (i=0;i<n;i++) sprintf(&b[strlen(b)]," %10.4lf",v[i]);
	rob_kin_user_msg(b);

	sprintf(b,"achs_rotx");
	v = inq_achs_rotx();
	for (i=0;i<n;i++) sprintf(&b[strlen(b)]," %10.4lf",v[i]*DEG);
	rob_kin_user_msg(b);

	sprintf(b,"achs_roty");
	v = inq_achs_roty();
	for (i=0;i<n;i++) sprintf(&b[strlen(b)]," %10.4lf",v[i]*DEG);
	rob_kin_user_msg(b);

	sprintf(b,"achs_rotz");
	v = inq_achs_rotz();
	for (i=0;i<n;i++) sprintf(&b[strlen(b)]," %10.4lf",v[i]*DEG);
	rob_kin_user_msg(b);

	sprintf(b,"Tp");
	v = (double *)T->p;
	for (i=0;i<3;i++) sprintf(&b[strlen(b)]," %10.4lf",v[i]);
	rob_kin_user_msg(b);

	rob_kin_user_msg("-------------------------------");

	return(WARN_OK);
} /* End of inv_kin_user_1*/

DllExport int  inv_kin_user_3(frame *bTt)
{
  double *achs_length,*achs_offsets1,*achs_offsets2,
		 *sing_tol,*joint_offset;
  float *q;

  int *warnings,*verpolt;
  int num_dofs,num_configs;
  frame *T,b0Tt;

  int i,ii, jj, kk, soln, warning;
  double nx,ny,nz,ox,oy,oz,ax, ay, az,px, py, pz;
  float l1x, l1y, l1z, l2x, l2y, l2z, l2, l3x, l3y, l3z, l3, l6,
		 ly, rx, rxy, rxm, rr;
  double c1, s1, c2, c3, c23, s23,s6,c6,s4,c4, cabs;
  double theta[KIN_DOFS];
  double delta_theta1, delta_theta2, delta_theta3;
  float sig1, sig2, sig3, sing_tolx, sing_tol_q5;

  float q_ext[KIN_DOFS];

//rob_kin_user_msg("Hallo, try this one");

	 num_dofs = inq_num_dofs();
	 num_configs = inq_num_configs();

	 if (num_dofs!=NUM_DOFS_3 || num_configs!=NUM_CONFIGS_3) {
		 char b[80];
		 rob_kin_user_msg("ERROR inv_kin_user_2()");
		 sprintf(b,"dofs    %2d != %2d",num_dofs,NUM_DOFS_3);
		 rob_kin_user_msg(b);
		 sprintf(b,"configs %2d != %2d",num_configs,NUM_CONFIGS_3);
		 rob_kin_user_msg(b);
	 }

	 achs_length = inq_achs_length();
	 achs_offsets1 = inq_achs_offsets1();
	 achs_offsets2 = inq_achs_offsets2();
	 joint_offset = inq_joint_offset();
	 sing_tol = inq_sing_tol();
	 warnings = inq_warnings();
	 verpolt = inq_verpolt();

	 l1z = achs_length[0];
	 l1x = achs_offsets1[0];
	 l1y = achs_offsets2[0];

	 l2z = achs_length[1];
	 l2x = achs_offsets1[1];
	 l2y = achs_offsets2[1];

	 if(fabs(l2x)>0.0 && fabs(l2z)>0.0) {
		 delta_theta2 = atan(l2x/l2z);
		 l2  = sqrt( l2x*l2x + l2z*l2z );
	 } else {
		 delta_theta2 = 0.0;
		 l2 = l2z;
	 }

	 l3z = achs_length[2];
	 l3x = achs_offsets1[2];
	 l3y = achs_offsets2[2];

	 if(fabs(l3x)>0.0 && fabs(l3z)>0.0) {
		 delta_theta3 = atan(l3x/l3z);
		 l3  = sqrt( l3x*l3x + l3z*l3z );
	 } else {
		 delta_theta3 = 0.0;
		 l3 = l3z;
	 }

	 ly = l1y + l2y + l3y;

	 l6 = achs_length[5];

	 sing_tolx = sing_tol[0]; /* [mm] */
	 sing_tol_q5 = (float)sin(sing_tol[4]); /* [rad] */

	 T = &b0Tt;
	 mul_invT_T(T,inq_bT0(),bTt);

	 px = T->p[0]-T->a[0]*l6;
	 py = T->p[1]-T->a[1]*l6;
	 pz = T->p[2]-T->a[2]*l6-l1z;

	 rxy = sqrt(px*px+py*py);

	 if( rxy < ly ) {
		  rob_kin_set_warnings(WARN_UNREACH);
		  return( WARN_UNREACH );
	 } else if( rxy < sing_tolx ) {
		  rob_kin_set_warnings(WARN_SINGULAR);
	 }

	 if (fabs(ly)>0.0) {
		 rx = rxy*rxy - ly*ly;
		 if (rx<0.0) rx=0.0;
		 else rx = sqrt( rx );
		 delta_theta1 = rob_kin_atan2(ly,rx);
	 } else {
		 rx = rxy;
		 delta_theta1 = 0.0;
	 }

	 sig1 = -1.0;
	 sig2 = -1.0;
	 soln = 0;

	// save values from external joints
    q = inq_q_solut();
    //moni_v("q",q,num_dofs,1.0);
	for( i = 6; i < num_dofs; i++ ) q_ext[i] = q[i];

	for( kk = 0; kk < 2; kk++ )
	{

	 warning=WARN_OK;

	 sig1 *= -1.0;

	 rxm = rx - sig1*l1x;

	 rr = sqrt( rxm*rxm + pz*pz );

	 if (rr<KIN_EPSI) {
		  rob_kin_set_warnings(WARN_UNREACH);
		  return( WARN_UNREACH );
	 }

	 c2 = (rr*rr + l2*l2 - l3*l3) / (2.0*l2*rr);

	 if( (cabs=fabs(c2)) > 1.0 ) {
		if (cabs >1.0+KIN_EPSI) {
		  warning = WARN_UNREACH;
		}
		if (c2>1.0) c2 = 1.0;
		else c2 = -1.0;
	 }

	 c3 = (l3*l3 + l2*l2 - rr*rr) / (2.0*l2*l3);

	 if( (cabs=fabs(c3)) > 1.0 ) {
		if (cabs >1.0+KIN_EPSI) {
		  warning = WARN_UNREACH;
		}
		if (c3>1.0) c3 = 1.0;
		else c3 = -1.0;
	 }

	if (warning==WARN_OK) {
		theta[0] = -sig1*delta_theta1 +
					 rob_kin_atan2(sig1*py,sig1*px);
	}
	else {
		for (i=kk*4;i<(kk+1)*4;i++) warnings[i] = WARN_UNREACH;
		soln = (kk+1)*4;
	}

	for( jj = 0; jj < 2 && warning==WARN_OK; jj++ ) {

		 sig2 *= -1.0;

		 theta[1] = sig1*(PI/2.0-rob_kin_atan2(pz,rxm)-sig2*acos(c2))
					 - delta_theta2;

		 if (theta[1] < -PI) theta[1] += 2*PI;
		 else if (theta[1] > PI)  theta[1] -= 2*PI;

		 theta[2] = -sig1*sig2*acos( c3 )
					 - delta_theta3 + delta_theta2;

		 if( theta[2] < 0.0 ) theta[2] += PI;
		 else theta[2] -= PI;

		 c1 = cos( theta[0] );
		 s1 = sin( theta[0] );

		 c23 = cos( theta[2] + theta[1] );
		 s23 = sin( theta[2] + theta[1] );

		 nx = (T->n[0]*c1+T->n[1]*s1)*c23-T->n[2]*s23;
		 ny = -T->n[0]*s1+T->n[1]*c1;
		 nz = (T->n[0]*c1+T->n[1]*s1)*s23+T->n[2]*c23;

		 ox = (T->o[0]*c1+T->o[1]*s1)*c23-T->o[2]*s23;
		 oy = -T->o[0]*s1+T->o[1]*c1;
		 oz = (T->o[0]*c1+T->o[1]*s1)*s23+T->o[2]*c23;

		 ax = (T->a[0]*c1+T->a[1]*s1)*c23-T->a[2]*s23;
		 ay = -T->a[0]*s1+T->a[1]*c1;
		 az = (T->a[0]*c1+T->a[1]*s1)*s23+T->a[2]*c23;

		 sig3 = -1.0;

		 for( ii = 0; ii < 2 ; ii++ ) {

			 sig3 *= -1.0;

			 theta[5] = rob_kin_atan2(  sig3*oz,-sig3*nz);
			 c6 = cos(theta[5]); s6 = sin(theta[5]);
			 theta[3] = rob_kin_atan2(-ox*c6-nx*s6,oy*c6+ny*s6);
			 c4 = cos(theta[3]); s4 = sin(theta[3]);
			 theta[4] = rob_kin_atan2(ax*c4+ay*s4,az);

			 if( fabs( sin(theta[4]) ) < sing_tol_q5 )
			 {
				 rob_kin_user_msg("inv_kin_user_2 WARN_SINGULAR");
				 warnings[soln] = WARN_SINGULAR;
			 }
			 else warnings[soln] = WARN_OK;

			 // copy values from external joints back
			 for( i = 0; i < NUM_DOFS_3; i++ ) {
				 q = inq_q(soln);
				 if (i<6) {
					 q[i] =  verpolt[i]*(theta[i]
										+joint_offset[i]);
				 } else {
					 q[i] =  q_ext[i];
				 }

				 if (q[i] < -PI) q[i] += 2*PI;
				 else if (q[i] > PI)  q[i] -= 2*PI;
			 }
			 soln++;
		 } /* for ii */
	} /* for jj */
	} /* for kk */
	return(WARN_OK);
} /* End of inv_kin_user_3*/

DllExport int  inv_kin_user_4(frame *T)
{
	moni_msg("inv_kin_user_4 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_4*/

DllExport int  inv_kin_user_5(frame *T)
{
	moni_msg("inv_kin_user_5 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_5*/

DllExport int  inv_kin_user_6(frame *T)
{
	moni_msg("inv_kin_user_6 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_6*/

DllExport int  inv_kin_user_7(frame *T)
{
	moni_msg("inv_kin_user_7 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_7*/

DllExport int  inv_kin_user_8(frame *T)
{
	moni_msg("inv_kin_user_8 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_8*/

DllExport int  inv_kin_user_9(frame *T)
{
	moni_msg("inv_kin_user_9 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_9*/

DllExport int  inv_kin_user_10(frame *T)
{
	moni_msg("inv_kin_user_10 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_10*/

DllExport int  inv_kin_user_11(frame *T)
{
	moni_msg("inv_kin_user_11 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_11*/

DllExport int  inv_kin_user_12(frame *T)
{
	moni_msg("inv_kin_user_12 x=%f",T->p[0]);
	return(WARN_OK);
} /* End of inv_kin_user_12*/


/*----------------------------------------
	Mathematical Dependency Functions
----------------------------------------*/
DllExport float  md_f1(void)
/*
	example for robot uni_tmpl.rob with 1 DOF
	for the first passiv joint
*/
{
	float *q = inq_q_solut();
	return(2*q[0]);
}
DllExport float  md_f2(void)
/*
	example for robot uni_tmpl.rob with 1 DOF
	for the 3rd passiv joint
*/
{
	float v,*q = inq_q_solut();
	v = (float)sin((double)q[0])+
		 (float)cos((double)q[0]);
	return(v);
}
#define MR 1.0
DllExport float  md_f3(void)
{
//	rob_kin_user_msg("md_f3");
	float *q = inq_q_solut(),phi;
		  phi = q[0]/MR;
	return(MR*(float)sin((double)phi));
}
DllExport float  md_f4(void)
{
//	rob_kin_user_msg("md_f4");
	float *q = inq_q_solut(),phi;
		  phi = q[0]/MR;
	return(MR*(float)(1-cos((double)phi)));
}
DllExport float  md_f5(void)
{
//	rob_kin_user_msg("md_f5");
	float *q = inq_q_solut(),phi;
		  phi = q[0]/MR;
	return(-phi);
}
DllExport float  md_f6(void)
{
	rob_kin_user_msg("md_f6");
	return(0);
}
DllExport float  md_f7(void)
{
	rob_kin_user_msg("md_f7");
	return(0);
}
DllExport float  md_f8(void)
{
	rob_kin_user_msg("md_f8");
	return(0);
}
// Rohr
#define D 0.4 // Durchmesser Auen
#define d 0.02 // Dicke des Rohres
#define da 0.32 // Durchmesser des Loches
#define R (0.5*D-d) // Radius Innen
DllExport float  md_f9(void)
{
//	rob_kin_user_msg("md_f9");
	// rohr, alfa
	float *q = inq_q_solut(),phi,alfa;
	phi = sssa(da,R,R);
	alfa = 0.5*phi*sin(q[3]);
	//moni_msg("alfa %f phi %f q4 %f R %f da %f",alfa*DEG,phi*DEG,q[3]*DEG,R,da);
	//mp=phi;
	//ma=alfa;
	return(alfa);
}
DllExport float  md_f10(void)
{
//	rob_kin_user_msg("md_f10");
	float *q = inq_q_solut(),x;
	x = 0.5*da*cos(q[3]);
	//moni_msg("a %f x %f q4 %f p %f",ma*DEG,x,q[3]*DEG,mp*DEG);
	return(x);
}
DllExport float  md_f11(void)
{
	rob_kin_user_msg("md_f11");
	return(0);
}
DllExport float  md_f12(void)
{
	rob_kin_user_msg("md_f12");
	return(0);
}

